from openshift.dynamic.apply import merge, apply_patch

tests = [
    dict(
        last_applied = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", two="2")
        ),
        desired = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", two="2")
        ),
        expected = {}
    ),
    dict(
        last_applied = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", two="2")
        ),
        desired = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", two="2", three="3")
        ),
        expected = dict(data=dict(three="3"))
    ),
    dict(
        last_applied = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", two="2")
        ),
        desired = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", three="3")
        ),
        expected = dict(data=dict(two=None, three="3"))
    ),
    dict(
        last_applied = dict(
            kind="ConfigMap",
            metadata=dict(name="foo", annotations=dict(this="one", hello="world")),
            data=dict(one="1", two="2")
        ),
        desired = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", three="3")
        ),
        expected = dict(metadata=dict(annotations=None), data=dict(two=None, three="3"))
    ),

    dict(
        last_applied = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        actual = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, protocol='TCP', name="http")])
        ),
        desired = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        expected = dict(spec=dict(ports=[dict(port=8080, protocol='TCP', name="http")]))
    ),
    dict(
        last_applied = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        actual = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, protocol='TCP', name="http")])
        ),
        desired = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8081, name="http")])
        ),
        expected = dict(spec=dict(ports=[dict(port=8081, name="http")]))
    ),
    dict(
        last_applied = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        actual = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, protocol='TCP', name="http")])
        ),
        desired = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8443, name="https"), dict(port=8080, name="http")])
        ),
        expected = dict(spec=dict(ports=[dict(port=8443, name="https"), dict(port=8080, name="http", protocol='TCP')]))
    ),
    dict(
        last_applied = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8443, name="https"), dict(port=8080, name="http")])
        ),
        actual = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8443, protocol='TCP', name="https"), dict(port=8080, protocol='TCP', name='http')])
        ),
        desired = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        expected = dict(spec=dict(ports=[dict(port=8080, name="http", protocol='TCP')]))
    ),
    dict(
        last_applied = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8443, name="https", madeup="xyz"), dict(port=8080, name="http")])
        ),
        actual = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8443, protocol='TCP', name="https", madeup="xyz"), dict(port=8080, protocol='TCP', name='http')])
        ),
        desired = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8443, name="https")])
        ),
        expected = dict(spec=dict(ports=[dict(madeup=None, port=8443, name="https", protocol='TCP')]))
    ),
    dict(
        last_applied = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox",
                                       resources=dict(requests=dict(cpu="100m", memory="100Mi"), limits=dict(cpu="100m", memory="100Mi")))])
        ),
        actual = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox",
                                       resources=dict(requests=dict(cpu="100m", memory="100Mi"), limits=dict(cpu="100m", memory="100Mi")))])
        ),
        desired = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox",
                                       resources=dict(requests=dict(cpu="50m", memory="50Mi"), limits=dict(memory="50Mi")))])
        ),
        expected=dict(spec=dict(containers=[dict(name="busybox", image="busybox",
                                                 resources=dict(requests=dict(cpu="50m", memory="50Mi"), limits=dict(cpu=None, memory="50Mi")))]))
    ),
    dict(
        desired = dict(kind='Pod',
                       spec=dict(containers=[
                           dict(name='hello',
                                volumeMounts=[dict(name="test", mountPath="/test")])
                           ],
                           volumes=[
                               dict(name="test", configMap=dict(name="test")),
                           ])),
        last_applied = dict(kind='Pod',
                       spec=dict(containers=[
                           dict(name='hello',
                                volumeMounts=[dict(name="test", mountPath="/test")])
                           ],
                           volumes=[
                               dict(name="test", configMap=dict(name="test")),
                           ])),
        actual = dict(kind='Pod',
                       spec=dict(containers=[
                           dict(name='hello',
                                volumeMounts=[dict(name="test", mountPath="/test"),
                                              dict(mountPath="/var/run/secrets/kubernetes.io/serviceaccount", name="default-token-xyz")])
                           ],
                           volumes=[
                               dict(name="test", configMap=dict(name="test")),
                               dict(name="default-token-xyz", secret=dict(secretName="default-token-xyz")),
                           ])),
        expected = dict(spec=dict(containers=[
                           dict(name='hello',
                                volumeMounts=[dict(name="test", mountPath="/test"),
                                              dict(mountPath="/var/run/secrets/kubernetes.io/serviceaccount", name="default-token-xyz")])
                           ],
                           volumes=[
                               dict(name="test", configMap=dict(name="test")),
                               dict(name="default-token-xyz", secret=dict(secretName="default-token-xyz")),
                           ])),
    ),

    # This next one is based on a real world case where definition was mostly
    # str type and everything else was mostly unicode type (don't ask me how)
    dict(
        last_applied = {
            u'kind': u'ConfigMap',
            u'data': {u'one': '1', 'three': '3', 'two': '2'},
            u'apiVersion': u'v1',
            u'metadata': {u'namespace': u'apply', u'name': u'apply-configmap'}
        },
        actual = {
            u'kind': u'ConfigMap',
            u'data': {u'one': '1', 'three': '3', 'two': '2'},
            u'apiVersion': u'v1',
            u'metadata': {u'namespace': u'apply', u'name': u'apply-configmap',
                          u'resourceVersion': '1714994',
                          u'creationTimestamp': u'2019-08-17T05:08:05Z', u'annotations': {},
                          u'selfLink': u'/api/v1/namespaces/apply/configmaps/apply-configmap',
                          u'uid': u'fed45fb0-c0ac-11e9-9d95-025000000001'}
        },
        desired = {
            'kind': u'ConfigMap',
            'data': {'one': '1', 'three': '3', 'two': '2'},
            'apiVersion': 'v1',
            'metadata': {'namespace': 'apply', 'name': 'apply-configmap'}
        },
        expected = dict()
    ),
    # apply a Deployment, then scale the Deployment (which doesn't affect last-applied)
    # then apply the Deployment again. Should un-scale the Deployment
    dict(
        last_applied = {
            'kind': u'Deployment',
            'spec': {
                'replicas': 1,
                'template': {
                    'spec': {
                        'containers': [
                            {
                                'name': 'this_must_exist',
                                'envFrom': [
                                    {
                                        'configMapRef': {
                                            'name': 'config-xyz'
                                        }
                                    },
                                    {
                                        'secretRef': {
                                            'name': 'config-wxy'
                                        }
                                    }
                                ]
                            }
                        ]
                    }
                }
            },
            'metadata': {
                'namespace': 'apply',
                'name': u'apply-deployment'
            }
        },
        actual = {
            'kind': u'Deployment',
            'spec': {
                'replicas': 0,
                'template': {
                    'spec': {
                        'containers': [
                            {
                                'name': 'this_must_exist',
                                'envFrom': [
                                    {
                                        'configMapRef': {
                                            'name': 'config-xyz'
                                        }
                                    },
                                    {
                                        'secretRef': {
                                            'name': 'config-wxy'
                                        }
                                    }
                                ]
                            }
                        ]
                    }
                }
            },
            'metadata': {
                'namespace': 'apply',
                'name': u'apply-deployment'
            }
        },
        desired = {
            'kind': u'Deployment',
            'spec': {
                'replicas': 1,
                'template': {
                    'spec': {
                        'containers': [
                            {
                                'name': 'this_must_exist',
                                'envFrom': [
                                    {
                                        'configMapRef': {
                                            'name': 'config-abc'
                                        }
                                    }
                                ]
                            }
                        ]
                    }
                }
            },
            'metadata': {
                'namespace': 'apply',
                'name': u'apply-deployment'
            }
        },
        expected = {
            'spec' : {
                'replicas': 1,
                'template': {
                    'spec': {
                        'containers': [
                            {
                                'name': 'this_must_exist',
                                'envFrom': [
                                    {
                                        'configMapRef': {
                                            'name': 'config-abc'
                                        }
                                    }
                                ]
                            }
                        ]
                    }
                }
            }
        }
    ),
    dict(
        last_applied = {
            'kind': 'MadeUp',
            'toplevel': {
                'original': 'entry'
            }
        },
        actual = {
            'kind': 'MadeUp',
            'toplevel': {
                'original': 'entry',
                'another': {
                    'nested': {
                         'entry': 'value'
                    }
                }
            }
        },
        desired = {
            'kind': 'MadeUp',
            'toplevel': {
                'original': 'entry',
                'another': {
                    'nested': {
                         'entry': 'value'
                    }
                }
            }
        },
        expected = {}
    )
]


def test_merges():
    for test in tests:
        assert(merge(test['last_applied'], test['desired'], test.get('actual', test['last_applied'])) == test['expected'])


def test_apply_patch():
        actual = dict(
            kind="ConfigMap",
            metadata=dict(name="foo",
                          annotations={'kubectl.kubernetes.io/last-applied-configuration':
                                       '{"data":{"one":"1","two":"2"},"kind":"ConfigMap",'
                                       '"metadata":{"annotations":{"hello":"world","this":"one"},"name":"foo"}}',
                                       'this': 'one', 'hello': 'world'}),
            data=dict(one="1", two="2")
        )
        desired = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(one="1", three="3")
        )
        expected = dict(
            metadata=dict(
                annotations={'kubectl.kubernetes.io/last-applied-configuration': '{"data":{"one":"1","three":"3"},"kind":"ConfigMap","metadata":{"name":"foo"}}',
                             'this': None, 'hello': None}),
                data=dict(two=None, three="3"))
        assert(apply_patch(actual, desired) == (actual, expected))



import time
from datetime import datetime

import requests

from . import RestClientTestCase

class TestOpenshiftApis(RestClientTestCase):

    def create_hello_openshift(self):
        apps_v1_deployments = self.client.resources.get(api_version='apps/v1', kind='Deployment')
        v1_services = self.client.resources.get(api_version='v1', kind='Service')

        name = 'hello-openshift'

        deployment = {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "name": name,
                "namespace": self.namespace
            },
            "spec": {
                "replicas": 3,
                "selector": {
                    "matchLabels": {
                        "app": "hello-openshift"
                    }
                },
                "template": {
                    "metadata": {
                        "labels": {
                            "app": "hello-openshift"
                        }
                    },
                    "spec": {
                        "containers": [{
                            "name": "hello-openshift",
                            "image": "docker.io/openshift/hello-openshift",
                            "ports": [{"containerPort": 8080}]
                        }]
                    }}}}
        service = {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "name": name,
                "namespace": self.namespace
            },
            "spec": {
                "ports": [{
                    "port": 80,
                    "targetPort": 8080
                }],
                "selector": {
                    "app": "hello-openshift"
                }
            }}
        self.addCleanup(apps_v1_deployments.delete, name=name, namespace=self.namespace)
        apps_v1_deployments.create(deployment)

        self.addCleanup(v1_services.delete, name=name, namespace=self.namespace)
        v1_services.create(service)

        # Wait 1 minute for deployment to become available
        timeout = 60
        start = datetime.now()
        while (datetime.now() - start).seconds < timeout:
            try:
                deployment = apps_v1_deployments.get(name=name, namespace=self.namespace)
                if (deployment.status
                    and deployment.spec.replicas == (deployment.status.replicas or 0)
                    and deployment.status.readyReplicas == deployment.spec.replicas
                    and deployment.status.observedGeneration == deployment.metadata.generation
                    and not deployment.status.unavailableReplicas):
                    return
                time.sleep(1)
            except Exception:
                time.sleep(1)

    def test_v1_route(self):
        self.create_hello_openshift()

        v1_routes = self.client.resources.get(api_version='route.openshift.io/v1', kind='Route')
        route = {
            "apiVersion": "route.openshift.io/v1",
            "kind": "Route",
            "metadata": {
                "name": "test-route",
                "namespace": self.namespace,
            },
            "spec": {
                "to": {
                    "kind": "Service",
                    "name": "hello-openshift",
                },
                "port": {
                    "targetPort": 8080
                },
                "tls": {
                    "termination": "Edge",
                    "insecureEdgeTerminationPolicy": "Redirect",
                }
            }
        }

        self.addCleanup(v1_routes.delete, name='test-route', namespace=self.namespace)
        created_route = v1_routes.create(route)

        url = created_route.spec.host
        timeout = 10
        start = datetime.now()
        while (datetime.now() - start).seconds < timeout:
            response = requests.get("https://{0}".format(url), verify=False)
            if response.status_code == 200:
                break
        assert response.text == "Hello OpenShift!\n"

    def test_templates(self):
        v1_templates = self.client.resources.get(api_version='template.openshift.io/v1', name='templates')
        v1_processed_templates = self.client.resources.get(api_version='template.openshift.io/v1', name='processedtemplates')

        nginx_template = v1_templates.get(name='nginx-example', namespace='openshift').to_dict()
        nginx_template = self.update_template_param(nginx_template, 'NAMESPACE', self.namespace)
        nginx_template = self.update_template_param(nginx_template, 'NAME', 'test123')

        response = v1_processed_templates.create(body=nginx_template, namespace=self.namespace)

        for obj in response.objects:
            if obj.metadata.namespace:
                assert obj.metadata.namespace == self.namespace
            assert obj.metadata.name == 'test123'

    def update_template_param(self, template, k, v):
        for i, param in enumerate(template['parameters']):
            if param['name'] == k:
                template['parameters'][i]['value'] = v
                return template
        return template



import pytest

from kubernetes.client import ApiClient

from openshift.dynamic import (
    DynamicClient,
    Resource,
    ResourceList,
    EagerDiscoverer,
    LazyDiscoverer
)


@pytest.fixture(scope='module')
def mock_namespace():
    return Resource(
        api_version='v1',
        kind='Namespace',
        name='namespaces',
        namespaced=False,
        preferred=True,
        prefix='api',
        shorter_names=['ns'],
        shortNames=['ns'],
        singularName='namespace',
        verbs=['create', 'delete', 'get', 'list', 'patch', 'update', 'watch']
    )

@pytest.fixture(scope='module')
def mock_templates():
    return Resource(
        api_version='v1',
        kind='Template',
        name='templates',
        namespaced=True,
        preferred=True,
        prefix='api',
        shorter_names=[],
        shortNames=[],
        verbs=['create', 'delete', 'get', 'list', 'patch', 'update', 'watch']
    )

@pytest.fixture(scope='module')
def mock_processedtemplates():
    return Resource(
        api_version='v1',
        kind='Template',
        name='processedtemplates',
        namespaced=True,
        preferred=True,
        prefix='api',
        shorter_names=[],
        shortNames=[],
        verbs=['create', 'delete', 'get', 'list', 'patch', 'update', 'watch']
    )

@pytest.fixture(scope='module')
def mock_namespace_list(mock_namespace):
    ret = ResourceList(mock_namespace.client, mock_namespace.group, mock_namespace.api_version, mock_namespace.kind)
    ret._ResourceList__base_resource = mock_namespace
    return ret


@pytest.fixture(scope='function', autouse=True)
def setup_client_monkeypatch(monkeypatch, mock_namespace, mock_namespace_list, mock_templates, mock_processedtemplates):

    def mock_load_server_info(self):
        self.__version = {'kubernetes': 'mock-k8s-version'}

    def mock_parse_api_groups(self, request_resources=False):
            return {
                'api': {
                    '': {
                        'v1': {
                            'Namespace': [mock_namespace],
                            'NamespaceList': [mock_namespace_list],
                            'Template': [mock_templates, mock_processedtemplates],
                        }
                    }
                }
            }

    monkeypatch.setattr(EagerDiscoverer, '_load_server_info', mock_load_server_info)
    monkeypatch.setattr(EagerDiscoverer, 'parse_api_groups', mock_parse_api_groups)
    monkeypatch.setattr(LazyDiscoverer, '_load_server_info', mock_load_server_info)
    monkeypatch.setattr(LazyDiscoverer, 'parse_api_groups', mock_parse_api_groups)


@pytest.fixture(params=[EagerDiscoverer, LazyDiscoverer])
def client(request):
    return DynamicClient(ApiClient(), discoverer=request.param)


@pytest.mark.parametrize(("attribute", "value"), [
    ('name', 'namespaces'),
    ('singular_name', 'namespace'),
    ('short_names', ['ns'])
])
def test_search_returns_single_and_list(client, mock_namespace, mock_namespace_list, attribute, value):
    resources = client.resources.search(**{'api_version':'v1', attribute: value})

    assert len(resources) == 2
    assert mock_namespace in resources
    assert mock_namespace_list in resources


@pytest.mark.parametrize(("attribute", "value"), [
    ('kind', 'Namespace'),
    ('name', 'namespaces'),
    ('singular_name', 'namespace'),
    ('short_names', ['ns'])
])
def test_get_returns_only_single(client, mock_namespace, attribute, value):
    resource = client.resources.get(**{'api_version':'v1', attribute: value})

    assert resource == mock_namespace


def test_get_namespace_list_kind(client, mock_namespace_list):
    resource = client.resources.get(api_version='v1', kind='NamespaceList')

    assert resource == mock_namespace_list


def test_search_multiple_resources_for_template(client, mock_templates, mock_processedtemplates):
    resources = client.resources.search(api_version='v1', kind='Template')

    assert len(resources) == 2
    assert mock_templates in resources
    assert mock_processedtemplates in resources



import unittest

from kubernetes import config
from kubernetes.client import api_client, Configuration

from openshift.dynamic import DynamicClient

class RestClientTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        try:
            config.load_kube_config()
        except config.ConfigException:
            config.load_incluster_config()
        cls.config = Configuration.get_default_copy()
        cls.client = DynamicClient(api_client.ApiClient(configuration=cls.config))

    def setUp(self):
        # Generate a namespace name from the test case name
        self.namespace = self.id().split('.')[-1].replace('_', '-')
        v1_namespaces = self.client.resources.get(api_version='v1', kind='Namespace')

        self.addCleanup(v1_namespaces.delete, name=self.namespace)
        v1_namespaces.create(body=dict(
            apiVersion='v1',
            kind='Namespace',
            metadata=dict(name=self.namespace),
        ))



# Test ConfigMap and Secret marshalling
# tests based on https://github.com/kubernetes/kubernetes/pull/49961

from openshift.helper.hashes import marshal, sorted_dict

tests = [
    dict(
        resource=dict(
            kind="ConfigMap",
            name="",
            data=dict(),
        ),
        expected=b'{"data":{},"kind":"ConfigMap","name":""}'
    ),
    dict(
        resource=dict(
            kind="ConfigMap",
            name="",
            data=dict(
                one=""
            ),
        ),
        expected=b'{"data":{"one":""},"kind":"ConfigMap","name":""}'
    ),
    dict(
        resource=dict(
            kind="ConfigMap",
            name="",
            data=dict(
                two="2",
                one="",
                three="3",
            ),
        ),
        expected=b'{"data":{"one":"","three":"3","two":"2"},"kind":"ConfigMap","name":""}'
    ),
    dict(
        resource=dict(
            kind="Secret",
            type="my-type",
            name="",
            data=dict(),
        ),
        expected=b'{"data":{},"kind":"Secret","name":"","type":"my-type"}'
    ),
    dict(
        resource=dict(
            kind="Secret",
            type="my-type",
            name="",
            data=dict(
                one=""
            ),
        ),
        expected=b'{"data":{"one":""},"kind":"Secret","name":"","type":"my-type"}'
    ),
    dict(
        resource=dict(
            kind="Secret",
            type="my-type",
            name="",
            data=dict(
                two="Mg==",
                one="",
                three="Mw==",
            ),
        ),
        expected=b'{"data":{"one":"","three":"Mw==","two":"Mg=="},"kind":"Secret","name":"","type":"my-type"}'
    ),
]

def test_marshal():
    for test in tests:
        assert(marshal(sorted_dict(test['resource']), sorted(list(test['resource'].keys()))) == test['expected'])



from openshift.dynamic.apply import recursive_diff

tests = [
    dict(
        before = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        after = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        expected = None
    ),
    dict(
        before = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http")])
        ),
        after = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8081, name="http")])
        ),
        expected = (
            dict(spec=dict(ports=[dict(port=8080, name="http")])),
            dict(spec=dict(ports=[dict(port=8081, name="http")]))
        )
    ),
    dict(
        before = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8080, name="http"), dict(port=8081, name="https")])
        ),
        after = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8081, name="https"), dict(port=8080, name="http")])
        ),
        expected = None
    ),

    dict(
        before = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox",
                                       env=[dict(name="hello", value="world"),
                                            dict(name="another", value="next")])])
        ),
        after = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox",
                                       env=[dict(name="hello", value="everyone")])])
        ),
        expected=(dict(spec=dict(containers=[dict(name="busybox", env=[dict(name="another", value="next"), dict(name="hello", value="world")])])),
                  dict(spec=dict(containers=[dict(name="busybox", env=[dict(name="hello", value="everyone")])])))
    ),

    dict(
        before = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox")])
        ),
        after = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8081, name="http")])
        ),
        expected=(dict(kind='Pod', spec=dict(containers=[dict(image='busybox', name='busybox')])),
                  dict(kind='Service', spec=dict(ports=[dict(name='http', port=8081)])))
    ),

    dict(
        before = dict(
            kind="Pod",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox")])
        ),
        after = dict(
            # kind="...",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8081, name="http")])
        ),
        expected=(dict(kind='Pod', spec=dict(containers=[dict(image='busybox', name='busybox')])),
                  dict(spec=dict(ports=[dict(name='http', port=8081)])))
    ),

    dict(
        before = dict(
            # kind="...",
            metadata=dict(name="foo"),
            spec=dict(containers=[dict(name="busybox", image="busybox")])
        ),
        after = dict(
            kind="Service",
            metadata=dict(name="foo"),
            spec=dict(ports=[dict(port=8081, name="http")])
        ),
        expected=(dict(spec=dict(containers=[dict(image='busybox', name='busybox')])),
                  dict(kind='Service', spec=dict(ports=[dict(name='http', port=8081)])))
    ),
    ]


def test_diff():
    for test in tests:
        assert(recursive_diff(test['before'], test['after']) == test['expected'])



from kubernetes.dynamic.resource import Resource, Subresource, ResourceField  # noqa


class ResourceList(Resource):
    """ Represents a list of API objects """

    def __init__(self, client, group='', api_version='v1', base_kind='', kind=None, base_resource_lookup=None):
        self.client = client
        self.group = group
        self.api_version = api_version
        self.kind = kind or '{}List'.format(base_kind)
        self.base_kind = base_kind
        self.base_resource_lookup = base_resource_lookup
        self.__base_resource = None

    def base_resource(self):
        if self.__base_resource:
            return self.__base_resource
        elif self.base_resource_lookup:
            self.__base_resource = self.client.resources.get(**self.base_resource_lookup)
            return self.__base_resource
        return None

    def apply(self, *args, **kwargs):
        return self.verb_mapper('apply', *args, **kwargs)

    def to_dict(self):
        return {
            '_type': 'ResourceList',
            'group': self.group,
            'api_version': self.api_version,
            'kind': self.kind,
            'base_kind': self.base_kind,
            'base_resource_lookup': self.base_resource_lookup
        }

    def __getattr__(self, name):
        if self.base_resource():
            return getattr(self.base_resource(), name)
        return None


class ResourceInstance(object):
    """ A parsed instance of an API resource. It exists solely to
        ease interaction with API objects by allowing attributes to
        be accessed with '.' notation.
    """

    def __init__(self, client, instance):
        self.client = client
        # If we have a list of resources, then set the apiVersion and kind of
        # each resource in 'items'
        kind = instance['kind']
        if kind.endswith('List') and 'items' in instance:
            kind = instance['kind'][:-4]
            if instance['items']:
                for item in instance['items']:
                    if 'apiVersion' not in item:
                        item['apiVersion'] = instance['apiVersion']
                    if 'kind' not in item:
                        item['kind'] = kind

        self.attributes = self.__deserialize(instance)
        self.__initialised = True



# Test ConfigMapHash and SecretHash equivalents
# tests based on https://github.com/kubernetes/kubernetes/pull/49961

from openshift.helper.hashes import generate_hash

tests = [
    dict(
        resource = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict()
        ),
        expected = "867km9574f",
    ),
    dict(
        resource = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            type="my-type",
            data=dict()
        ),
        expected = "867km9574f",
    ),
    dict(
        resource = dict(
            kind="ConfigMap",
            metadata=dict(name="foo"),
            data=dict(
                key1="value1",
                key2="value2")
        ),
        expected = "gcb75dd9gb",
    ),
    dict(
        resource = dict(
            kind="Secret",
            metadata=dict(name="foo"),
            data=dict()
        ),
        expected = "949tdgdkgg",
    ),
    dict(
        resource = dict(
            kind="Secret",
            metadata=dict(name="foo"),
            type="my-type",
            data=dict()
        ),
        expected = "dg474f9t76",
    ),

    dict(
        resource = dict(
            kind="Secret",
            metadata=dict(name="foo"),
            data=dict(
                key1="dmFsdWUx",
                key2="dmFsdWUy")
        ),
        expected = "tf72c228m4",
    )

]


def test_hashes():
    for test in tests:
        assert(generate_hash(test['resource']) == test['expected'])



from collections import OrderedDict
from copy import deepcopy
import json
import sys

from openshift.dynamic.exceptions import NotFoundError, ApplyException

LAST_APPLIED_CONFIG_ANNOTATION = 'kubectl.kubernetes.io/last-applied-configuration'

POD_SPEC_SUFFIXES = {
    'containers': 'name',
    'initContainers': 'name',
    'ephemeralContainers': 'name',
    'volumes': 'name',
    'imagePullSecrets': 'name',
    'containers.volumeMounts': 'mountPath',
    'containers.volumeDevices': 'devicePath',
    'containers.env': 'name',
    'containers.ports': 'containerPort',
    'initContainers.volumeMounts': 'mountPath',
    'initContainers.volumeDevices': 'devicePath',
    'initContainers.env': 'name',
    'initContainers.ports': 'containerPort',
    'ephemeralContainers.volumeMounts': 'mountPath',
    'ephemeralContainers.volumeDevices': 'devicePath',
    'ephemeralContainers.env': 'name',
    'ephemeralContainers.ports': 'containerPort',
}

POD_SPEC_PREFIXES = [
    'Pod.spec',
    'Deployment.spec.template.spec',
    'DaemonSet.spec.template.spec',
    'StatefulSet.spec.template.spec',
    'Job.spec.template.spec',
    'Cronjob.spec.jobTemplate.spec.template.spec',
]

# patch merge keys taken from generated.proto files under
# staging/src/k8s.io/api in kubernetes/kubernetes
STRATEGIC_MERGE_PATCH_KEYS = {
    'Service.spec.ports': 'port',
    'ServiceAccount.secrets': 'name',
    'ValidatingWebhookConfiguration.webhooks': 'name',
    'MutatingWebhookConfiguration.webhooks': 'name',
}

STRATEGIC_MERGE_PATCH_KEYS.update(
    {"%s.%s" % (prefix, key): value
     for prefix in POD_SPEC_PREFIXES
     for key, value in POD_SPEC_SUFFIXES.items()}
)


if sys.version_info.major >= 3:
    json_loads_byteified = json.loads
else:
    # https://stackoverflow.com/a/33571117
    def json_loads_byteified(json_text):
        return _byteify(
            json.loads(json_text, object_hook=_byteify),
            ignore_dicts=True
        )

    def _byteify(data, ignore_dicts = False):
        # if this is a unicode string, return its string representation
        if isinstance(data, unicode):  # noqa: F821
            return data.encode('utf-8')
        # if this is a list of values, return list of byteified values
        if isinstance(data, list):
            return [ _byteify(item, ignore_dicts=True) for item in data ]
        # if this is a dictionary, return dictionary of byteified keys and values
        # but only if we haven't already byteified it
        if isinstance(data, dict) and not ignore_dicts:
            return {
                _byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True)
                for key, value in data.items()
            }
        # if it's anything else, return it in its original form
        return data


def annotate(desired):
    return dict(
        metadata=dict(
            annotations={
                LAST_APPLIED_CONFIG_ANNOTATION: json.dumps(desired, separators=(',', ':'), indent=None, sort_keys=True)
            }
        )
    )


def apply_patch(actual, desired):
    last_applied = actual['metadata'].get('annotations',{}).get(LAST_APPLIED_CONFIG_ANNOTATION)

    if last_applied:
        # ensure that last_applied doesn't come back as a dict of unicode key/value pairs
        # json.loads can be used if we stop supporting python 2
        last_applied = json_loads_byteified(last_applied)
        patch = merge(dict_merge(last_applied, annotate(last_applied)),
                      dict_merge(desired, annotate(desired)), actual)
        if patch:
            return actual, patch
        else:
            return actual, actual
    else:
        return actual, dict_merge(desired, annotate(desired))


def apply_object(resource, definition):
    try:
        actual = resource.get(name=definition['metadata']['name'], namespace=definition['metadata'].get('namespace'))
    except NotFoundError:
        return None, dict_merge(definition, annotate(definition))
    return apply_patch(actual.to_dict(), definition)


def apply(resource, definition):
    existing, desired = apply_object(resource, definition)
    if not existing:
        return resource.create(body=desired, namespace=definition['metadata'].get('namespace'))
    if existing == desired:
        return resource.get(name=definition['metadata']['name'], namespace=definition['metadata'].get('namespace'))
    return resource.patch(body=desired,
                          name=definition['metadata']['name'],
                          namespace=definition['metadata'].get('namespace'),
                          content_type='application/merge-patch+json')


# The patch is the difference from actual to desired without deletions, plus deletions
# from last_applied to desired. To find it, we compute deletions, which are the deletions from
# last_applied to desired, and delta, which is the difference from actual to desired without
# deletions, and then apply delta to deletions as a patch, which should be strictly additive.
def merge(last_applied, desired, actual, position=None):
    deletions = get_deletions(last_applied, desired)
    delta = get_delta(last_applied, actual, desired, position or desired['kind'])
    return dict_merge(deletions, delta)


# dict_merge taken from Ansible's module_utils.common.dict_transformations
def dict_merge(a, b):
    '''recursively merges dicts. not just simple a['key'] = b['key'], if
    both a and b have a key whose value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary.'''
    if not isinstance(b, dict):
        return b
    result = deepcopy(a)
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
        else:
            result[k] = deepcopy(v)
    return result


def list_to_dict(lst, key, position):
    result = OrderedDict()
    for item in lst:
        try:
            result[item[key]] = item
        except KeyError:
            raise ApplyException("Expected key '%s' not found in position %s" % (key, position))
    return result


# list_merge applies a strategic merge to a set of lists if the patchMergeKey is known
# each item in the list is compared based on the patchMergeKey - if two values with the
# same patchMergeKey differ, we take the keys that are in last applied, compare the
# actual and desired for those keys, and update if any differ
def list_merge(last_applied, actual, desired, position):
    result = list()
    if position in STRATEGIC_MERGE_PATCH_KEYS and last_applied:
        patch_merge_key = STRATEGIC_MERGE_PATCH_KEYS[position]
        last_applied_dict = list_to_dict(last_applied, patch_merge_key, position)
        actual_dict = list_to_dict(actual, patch_merge_key, position)
        desired_dict = list_to_dict(desired, patch_merge_key, position)
        for key in desired_dict:
            if key not in actual_dict or key not in last_applied_dict:
                result.append(desired_dict[key])
            else:
                patch = merge(last_applied_dict[key], desired_dict[key], actual_dict[key], position)
                result.append(dict_merge(actual_dict[key], patch))
        for key in actual_dict:
            if key not in desired_dict and key not in last_applied_dict:
                result.append(actual_dict[key])
        return result
    else:
        return desired


def recursive_list_diff(list1, list2, position=None):
    result = (list(), list())
    if position in STRATEGIC_MERGE_PATCH_KEYS:
        patch_merge_key = STRATEGIC_MERGE_PATCH_KEYS[position]
        dict1 = list_to_dict(list1, patch_merge_key, position)
        dict2 = list_to_dict(list2, patch_merge_key, position)
        dict1_keys = set(dict1.keys())
        dict2_keys = set(dict2.keys())
        for key in dict1_keys - dict2_keys:
            result[0].append(dict1[key])
        for key in dict2_keys - dict1_keys:
            result[1].append(dict2[key])
        for key in dict1_keys & dict2_keys:
            diff = recursive_diff(dict1[key], dict2[key], position)
            if diff:
                # reinsert patch merge key to relate changes in other keys to
                # a specific list element
                diff[0].update({patch_merge_key: dict1[key][patch_merge_key]})
                diff[1].update({patch_merge_key: dict2[key][patch_merge_key]})
                result[0].append(diff[0])
                result[1].append(diff[1])
        if result[0] or result[1]:
            return result
    elif list1 != list2:
        return (list1, list2)
    return None


def recursive_diff(dict1, dict2, position=None):
    if not position:
        if 'kind' in dict1 and dict1.get('kind') == dict2.get('kind'):
            position = dict1['kind']
    left = dict((k, v) for (k, v) in dict1.items() if k not in dict2)
    right = dict((k, v) for (k, v) in dict2.items() if k not in dict1)
    for k in (set(dict1.keys()) & set(dict2.keys())):
        this_position = "%s.%s" % (position, k) if position is not None else None
        if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
            result = recursive_diff(dict1[k], dict2[k], this_position)
            if result:
                left[k] = result[0]
                right[k] = result[1]
        elif isinstance(dict1[k], list) and isinstance(dict2[k], list):
            result = recursive_list_diff(dict1[k], dict2[k], this_position)
            if result:
                left[k] = result[0]
                right[k] = result[1]
        elif dict1[k] != dict2[k]:
            left[k] = dict1[k]
            right[k] = dict2[k]
    if left or right:
        return left, right
    else:
        return None


def get_deletions(last_applied, desired):
    patch = {}
    for k, last_applied_value in last_applied.items():
        desired_value = desired.get(k)
        if isinstance(last_applied_value, dict) and isinstance(desired_value, dict):
            p = get_deletions(last_applied_value, desired_value)
            if p:
                patch[k] = p
        elif last_applied_value != desired_value:
            patch[k] = desired_value
    return patch


def get_delta(last_applied, actual, desired, position=None):
    patch = {}

    for k, desired_value in desired.items():
        if position:
            this_position = "%s.%s" % (position, k)
        actual_value = actual.get(k)
        if actual_value is None:
            patch[k] = desired_value
        elif isinstance(desired_value, dict):
            p = get_delta(last_applied.get(k, {}), actual_value, desired_value, this_position)
            if p:
                patch[k] = p
        elif isinstance(desired_value, list):
            p = list_merge(last_applied.get(k, []), actual_value, desired_value, this_position)
            if p:
                patch[k] = [item for item in p if item]
        elif actual_value != desired_value:
            patch[k] = desired_value
    return patch
